from logging import FileHandler
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import pandas as pd
import petname
from datasets import Dataset

from .src.common.logging import get_logger
from .src.common.utils import get_cache_path
from .src.config import Config, parse_metrics_config
from .src.data import (
    concatenate,
    domain_model_counts,
    errors_per_model,
    get_save_path,
    serialize_dataset,
)
from .src.exploration import get_explorer
from .src.generators import get_generator
from .src.metrics import run_metrics as _run_metrics
from .src.models.types import GENERATION_ERROR
from .src.postprocessing import filter_by_language, postprocess
from .src.types import TaskType

_logger = get_logger(__name__)


def generate_from_config(
    config: Config,
    save_dir: Path,
    run_name: str,
) -> Path:
    """
    Generates a dataset using `TextGeneration` parameterized
    by `config` and saves it.

    Args:
        config (Config): a configuration.
        save_dir (Path): root dir where to save the generated dataset.
        run_name (str): name of this run.

    Returns:
        Path: path where the generated dataset was saved.
    """
    generator = get_generator(config)
    dataset = generator.generate()

    errors = count_errors(dataset)
    _logger.info(f"{errors} errors found in the generated dataset.")

    dataset = filter_by_language(dataset, config.input.language)

    output_path = serialize_dataset(dataset, config, save_dir, run_name)

    return output_path


def _generate(
    config_path: Path,
    save_dir: Path,
    run_name: str,
    task_type: TaskType,
) -> None:
    """
    Runs the generation pipeline in an end-to-end manner.

    Args:
        config_path (Path): path containing YAML config files.
        save_dir (Path): root dir where to save the generated dataset.
        run_name (str): name of this run.
        task_type (TaskType): the type of task.
    """
    configs = Config.load_configs(config_path, task_type)

    _, statistics = generate_dataset(configs, save_dir, run_name)

    statistics_dir = save_dir / "statistics"
    statistics_dir.mkdir(parents=True, exist_ok=True)
    for name, df in statistics.items():
        df.to_markdown(statistics_dir / f"{name}.md")
        df.to_json(statistics_dir / f"{name}.json")


def generate_dataset(
    configs: List[Config],
    save_dir: Path,
    run_name: str,
) -> Tuple[Dataset, Dict[str, pd.DataFrame]]:
    """
    Generates a dataset given a list of configs.
    Only generates a dataset for a config if it hasn't
    been already generated for this `run_name`.

    Computes statistics for the generated dataset.

    Args:
        configs (List[Config]): list of configs to use for generation.
        save_dir (Path): root dir where to save the generated dataset.
        run_name (str): name of this run.
    Returns:
        Tuple[Dataset, Dict[str, pd.DataFrame]]: a tuple (dataset, statistics dict)
    """
    cache_path = get_cache_path()
    paths = []

    for config in configs:
        path = get_save_path(config, cache_path, run_name, check_exists=True)
        if not path:
            path = generate_from_config(config, cache_path, run_name)

        paths.append(path)

    dataset = concatenate(paths, save_dir)
    statistics = compute_statistics(dataset)
    errors = count_errors(dataset)

    dataset = postprocess(dataset, configs[0].task_type)

    dataset.save_to_disk(save_dir.as_posix())

    _logger.info(
        f"A total of {errors} errors have been found in the generation process."
    )
    _logger.info(f"Your dataset has been generated at {str(save_dir)}")

    return dataset, statistics


def compute_statistics(dataset: Dataset) -> Dict[str, pd.DataFrame]:
    """
    Computes a set of statistics of a generated dataset.

    Args:
        dataset (Dataset): the dataset of which statistics are computed.
    Returns:
        Dict[str, pd.DataFrame]: the statistics.
    """
    domain_model = domain_model_counts(dataset)
    errors = errors_per_model(dataset)

    return {"domain_model": domain_model, "errors": errors}


def generate_run_name() -> str:
    """
    Generates a name for a run.

    Returns:
        str: name of the run.
    """
    return petname.generate()


def log_final_message(run_name: str) -> None:
    """
    Logs the last logging message of TextMachina.

    Args:
        run_name (str): name of the run.
    """
    _logger.info(
        f"This run has been registered with name: '{run_name}'."
        " If the run terminated due to errors, you can use this"
        " name to continue from the where the process left off."
    )
    file_handler = [h for h in _logger.handlers if isinstance(h, FileHandler)][
        0
    ]
    _logger.info(f"Logfile saved in {file_handler.baseFilename}")


def count_errors(dataset: Dataset) -> int:
    """
    Counts the number of generation errors in the dataset.
    A generation error is identified as a text being `GENERATION_ERROR`.

    Args:
        dataset (Dataset): a dataset.

    Returns:
        int: number of errors in the dataset.
    """
    error_count = sum(
        [x in x for x in dataset["text"] if x == GENERATION_ERROR]
    )
    return error_count


def _explore(
    config_path: Path,
    metrics_path: Optional[Path],
    save_dir: Path,
    run_name: str,
    task_type: TaskType,
    interactive: bool,
    max_generations: int,
) -> None:
    """
    Carries out the exploration steps:
    - create a small dataset
    - generate a set of metrics based on the task type
    - interactive exploration through the dataset

    Args:
        config_path (Path): path containing YAML config files for generation.
        metrics_path (Optional[Path]): path to YAML config file of metrics.
        save_dir (Path): root dir where to save the generated dataset.
        run_name (str): name of this run.
        task_type (TaskType): the type of task to be explored.
        interactive(bool): whether to step interactively through the dataset.
        max_generations (int): the maximum number of texts to generate
            for each config. Ignored if the dataset has already been generated.
    """

    configs = Config.load_configs(
        config_path, task_type=task_type, max_generations=max_generations
    )

    dataset, _ = generate_dataset(configs, save_dir, run_name)

    if metrics_path:
        metrics, metric_args = parse_metrics_config(metrics_path)
        # Generate metrics
        if metrics:
            if metric_args is None:
                metric_args = {}
            _run_metrics(dataset, task_type, save_dir, metrics, metric_args)

    # Show examples in console
    if interactive:
        explorer = get_explorer(task_type, dataset)
        explorer.step()



from typing import Dict, List
from setuptools import setup, find_packages

VERSION: Dict[str, str] = {}

with open("text_machina/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()

TEXT_GENERATION_DEPS = [
    "scipy>=1.10.1",
    "PyYAML>=6.0.1",
    "datasets>=2.14.4",
    "spacy>=3.6.1",
    "typer>=0.9.0",
    "pydantic>=2.3.0",
    "petname>=2.6",
    "pycountry>=22.3.5",
    "ftfy>=6.1.3",
    "fasttext-wheel",
]

EXPLORE_DEPS = [
    "rich>=13.7.0",
    "scikit-learn>=1.3.2",
    "mauve-text>=0.3.0",
    "matplotlib>=3.7.4",
    "tabulate>=0.9.0",
    "readchar>=4.0.5",
    "evaluate>=0.4.1",
    "textstat>=0.7.3",
    "seqeval>=1.2.2",
]

DEPS = TEXT_GENERATION_DEPS + EXPLORE_DEPS

DEV_DEPS = [
    "black",
    "flake8",
    "mypy",
    "types-requests",
    "pytest",
    "isort",
    "autoflake",
    "pre-commit",
    "pytest-sphinx",
    "pytest-cov",
    "ruff",
    "Sphinx>=4.3.0,<7.1.0",
    "furo==2023.7.26",
    "myst-parser>=1.0,<2.1",
    "sphinx-copybutton==0.5.2",
    "sphinx-autobuild==2021.3.14",
    "sphinx-autodoc-typehints==1.23.3",
    "packaging",
    "setuptools",
    "build",
    "wheel",
]

EXTRAS_REQUIRES: Dict[str, List[str]] = {
    "openai": ["openai>=1", "tiktoken>=0.4.0"],
    "azure_openai": ["openai>=1", "tiktoken>=0.4.0"],
    "bedrock": ["boto3", "tiktoken>=0.4.0"],
    "ai21": ["ai21>=2.0.0", "ai21_tokenizer>=0.3.11"],
    "anthropic": ["anthropic>=0.7.2"],
    "cohere": ["cohere>=4.36"],
    "huggingface-local": [
        "torch==2.0.1",
        "transformers>=4.32.0",
        "accelerate>=0.22.0",
        "bitsandbytes>=0.41.1",
    ],
    "huggingface-remote": ["requests>=2.31.0"],
    "vertex": ["google-auth", "google-cloud-aiplatform==1.25.0", "tiktoken>=0.4.0"],
}
EXTRAS_REQUIRES["all"] = sum(EXTRAS_REQUIRES.values(), [])
EXTRAS_REQUIRES["dev"] = EXTRAS_REQUIRES["all"] + DEV_DEPS

setup(
    version=VERSION["VERSION"],
    name="text-machina",
    description="Text Machina: Seamless Generation of Machine-Generated Text Datasets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Genaios",
    entry_points={"console_scripts": ["text-machina=text_machina.cli:app"]},
    packages=find_packages(),
    install_requires=DEPS,
    include_package_data=True,
    python_requires=">=3.8.0",
    extras_require=EXTRAS_REQUIRES,
    url="https://github.com/Genaios/TextMachina",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
    ],
    license_files=[
        "LICENSE",
    ],
)



import os
from typing import List

import cohere

from .base import Tokenizer


class CohereTokenizer(Tokenizer):
    """
    Tokenizer for Cohere models.

    Requires the definition of the `COHERE_API_KEY=<key>` environment variable.
    """

    def __init__(self, model_name: str):
        self.model_name = model_name
        self.client = cohere.Client(api_key=os.environ["COHERE_API_KEY"])

    def encode(self, text: str) -> List[int]:
        return self.client.tokenize(text).tokens

    def decode(self, tokens: List[int]) -> str:
        return self.client.detokenize(tokens)



# flake8: noqa
from typing import Mapping, Type

from ..config import InputConfig
from ..types import TaskType
from .auxiliary import Auxiliary
from .base import Extractor
from .combined import Combined
from .dummy import Dummy
from .entity_list import EntityList
from .example import Example
from .noun_list import NounList
from .sentence_gap import SentenceGap
from .sentence_masking import SentenceMasking
from .sentence_prefix import SentencePrefix
from .sentence_rewriting import SentenceRewriting
from .word_gap import WordGap
from .word_masking import WordMasking
from .word_prefix import WordPrefix

EXTRACTORS: Mapping[str, Type[Extractor]] = {
    "auxiliary": Auxiliary,
    "combined": Combined,
    "dummy": Dummy,
    "entity_list": EntityList,
    "noun_list": NounList,
    "sentence_prefix": SentencePrefix,
    "sentence_gap": SentenceGap,
    "sentence_masking": SentenceMasking,
    "sentence_rewriting": SentenceRewriting,
    "word_prefix": WordPrefix,
    "word_gap": WordGap,
    "word_masking": WordMasking,
    "example": Example,
}


def get_extractor(
    extractor: str, input_config: InputConfig, task_type: TaskType
) -> Extractor:
    """
    Gets an extractor from the extractors pool.

    Args:
        extractor (str): extractor name.
        input_config (InputConfig): an input config.

    Returns:
        Extractor: an extractor from the pool.
    """
    return EXTRACTORS[extractor](input_config, task_type)


__all__ = [str(cls) for cls in EXTRACTORS.values()]



from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Type

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
)
from yaml import full_load, safe_load

from .types import TaskType


class InputConfig(BaseModel):
    """
    Wrapper for the input_config field.
    """

    quantity: int = Field(gt=0, description="Number of samples to generate.")
    domain: str = Field(description="Domain of a dataset.")
    dataset: str = Field(description="Name (HF Hub) or path to the dataset.")
    dataset_text_column: str = Field(
        description="Name of column in the dataset containing the text."
    )
    dataset_params: Dict[str, Any] = Field(
        description="Arguments to load the dataset."
    )
    template: str = Field(description="Template for the generations.")
    extractor: str = Field(description="Extractor name.")
    extractors_list: List[str] = Field(
        default=[],
        description=(
            "List of extractors to be used" " with the `combined` extractor."
        ),
        validate_default=True,
    )
    random_sample_human: bool = Field(
        default=False,
        desc=(
            "Whether to randomly sample human texts or use"
            " the same ones used to generate MGT"
        ),
    )
    max_input_tokens: int = Field(
        default=256,
        gt=0,
        desc=(
            "Maximum token length to be distributed across the"
            " prompt inputs extracted with the extractors."
        ),
    )
    extractor_args: Dict[str, Dict[str, Any]] = Field(
        default={}, desc="Extractors-specific arguments."
    )
    language: str = Field(
        default="en",
        desc="Language of the dataset used.",
        validate_default=True,
    )

    @field_validator("language")
    @classmethod
    def language_must_be_iso639(cls, language: str) -> str:
        import pycountry

        allowed_languages = [
            lang.alpha_2
            for lang in pycountry.languages
            if hasattr(lang, "alpha_2")
        ] + ["multilingual"]

        if language not in allowed_languages:
            from .common import InvalidLanguage

            raise InvalidLanguage()

        return language

    @field_validator("extractor")
    @classmethod
    def extractor_must_exist(cls, extractor: str) -> str:
        from .extractors import EXTRACTORS

        if extractor not in EXTRACTORS.keys():
            from .common import InvalidExtractor

            raise InvalidExtractor(extractor)

        return extractor

    @field_validator("extractors_list")
    @classmethod
    def not_empty_list_in_combined(
        cls, extractors_list: List[str], info: ValidationInfo
    ) -> List[str]:
        if info.data["extractor"] == "combined" and not extractors_list:
            from .common import CombinedEmptyExtractors

            raise CombinedEmptyExtractors()

        return extractors_list


class ModelConfig(BaseModel):
    """
    Wrapper for the input_config field.
    """

    provider: str = Field(description="Provider of text generation models.")
    model_name: str = Field(description="Name of a text generation model.")
    threads: int = Field(
        default=8,
        gt=0,
        description="Number of threads to use in `generate_completions`",
    )
    api_type: Literal["CHAT", "COMPLETION"] = Field(
        default="COMPLETION",
        description=(
            "API type for providers that allows chat and completion endpoints."
            "This arg must be `CHAT` or `COMPLETION` and must be according to"
            "the model used:\n"
            "- `CHAT`: for chat completion endpoints.\n"
            "- `COMPLETION` for traditional completion endpoints.\n"
            "For instance, GPT-4 in OpenAI can only be used with `CHAT`."
        ),
    )
    # Allow extra args and avoid protected naming conflicts
    model_config = ConfigDict(extra="allow", protected_namespaces=(""))

    @field_validator("provider")
    @classmethod
    def provider_must_exist(cls, provider: str) -> str:
        from .common import InvalidProvider
        from .models import MODELS

        if provider not in MODELS.keys():
            raise InvalidProvider(provider)
        return provider


class Config(BaseModel):
    """
    Wrapper for the config.
    """

    path: Optional[Path] = None
    task_type: TaskType
    input: InputConfig
    model: ModelConfig
    generation: Dict[str, Any]

    # Avoid protected naming conflicts
    model_config = ConfigDict(protected_namespaces=(""))

    @classmethod
    def load_config(
        cls: Type["Config"],
        path: Path,
        task_type: TaskType,
        max_generations: Optional[int] = None,
    ) -> "Config":
        config = safe_load(path.open("r"))

        if max_generations:
            config["input_config"]["quantity"] = min(
                max_generations, config["input_config"]["quantity"]
            )

        input = InputConfig(**config["input_config"])
        model = ModelConfig(**config["model_config"])
        generation = config["generation_config"]

        return cls(
            path=path,
            task_type=task_type,
            input=input,
            model=model,
            generation=generation,
        )

    @classmethod
    def load_configs(
        cls: Type["Config"],
        path: Path,
        task_type: TaskType,
        max_generations: Optional[int] = None,
    ) -> List["Config"]:
        path_iterator = (
            [path]
            if path.suffix in {".yml", ".yaml"}
            else chain(path.rglob("*.yml"), path.rglob("*.yaml"))
        )

        if not path_iterator:
            raise ValueError(
                "The provided path does not contain any YML files", path
            )

        return [
            cls.load_config(
                path=p, task_type=task_type, max_generations=max_generations
            )
            for p in path_iterator
        ]

    def safe_model_name(self) -> str:
        return self.model.model_name.split("/")[-1].replace("_", "-")

    def safe_dataset_name(self) -> str:
        return self.input.dataset.split("/")[-1].replace("_", "-")

    def safe_domain_name(self) -> str:
        return self.input.domain.replace("/", "-").replace("_", "-")


def parse_metrics_config(path: Path) -> Tuple[List[str], Dict]:
    """
    Parses a metrics config.

    Args:
        path (Path): the metric config path to parse.

    Returns:
        Tuple[List[str], Dict]: a tuple of structure (list of metrci names, args).
    """
    config = full_load(path.open("r"))
    metrics = config["metrics_to_run"]
    del config["metrics_to_run"]

    return metrics, config



import inspect
import os
from pathlib import Path
from typing import Callable, Dict, List, Optional


def get_instantiation_args(
    fn: Callable, args: Dict, accepted_params: Optional[List[str]] = None
) -> Dict:
    """
    Extracts the arguments from `args` that match the
    parameters of the `fn` function.

    Args:
        fn (Callable): a function.
        args (Dict): a dictionary of arguments.
        accepted_params (Optional[List[str]]): if known, the list
            of params accepted by fn. If `accepted_params` is
            provided, the `fn` argument will not be used to infer
            the accepted parameters.
    Returns:
        Dict: arguments in `args` that match the `fn` parameters.
    """
    if accepted_params is None:
        accepted_params = list(inspect.signature(fn).parameters.keys())
    return {arg: value for arg, value in args.items() if arg in accepted_params}


def get_cache_path() -> Path:
    """
    Get the TextMachina folder for caching intermediate results.

    Returns:
        Path: cache path.
    """
    path = Path(
        os.getenv("TEXT_MACHINA_CACHE_DIR", default="/tmp/text_machina_cache")
    )
    path.mkdir(parents=True, exist_ok=True)

    return path



from typing import List

import spacy
from tqdm import tqdm

from ..common import get_logger
from .types import SPACY_MODEL_MAPPING

_logger = get_logger(__name__)


def clean_inputs(texts: List[str]) -> List[str]:
    """
    Remove special symbols from the texts used as prompt inputs, to
    avoid breaking the format of classical kinds of prompts.

    Args:
        texts (List[str]): list of texts.
    Returns:
        List[str]: cleaned texts.
    """
    clean = []
    repl_map = {"\n": " ", "\t": " ", "\r": " ", "->": " ", ":": " "}
    for text in texts:
        for src, repl in repl_map.items():
            text = text.replace(src, repl)
        text = " ".join(text.split())
        text = text.strip()
        clean.append(text)
    return clean


def get_spacy_model(language: str) -> spacy.lang:
    """
    Gets or download a Spacy model.

    Args:
        language (str): language.

    Returns:
        spacy.lang: a Spacy model.
    """
    spacy_model = SPACY_MODEL_MAPPING.get(
        language, SPACY_MODEL_MAPPING["multilingual"]
    )

    try:
        nlp = spacy.load(spacy_model)
    except OSError:
        _logger.info(f"Downloading {spacy_model} from SpaCy.")
        spacy.cli.download(spacy_model)
        nlp = spacy.load(spacy_model)
    return nlp


def spacy_pipeline(
    texts: List[str],
    language: str,
    disable_pipes: List[str] = [],
    n_process: int = 4,
) -> List[spacy.tokens.Doc]:
    """
    Processes texts with spacy pipeline for entity extraction.

    Args:
        texts (List[str]): list of texts.
        language (str): language of the text.
        disable_pipes (List[str]): Spacy pipes to be disabled.
        n_process (int): number of processes.

    Returns:
        List[spacy.tokens.Doc]: list of Spacy docs.
    """
    nlp = get_spacy_model(language)

    processed_texts = list(
        tqdm(
            nlp.pipe(
                texts,
                n_process=n_process,
                disable=disable_pipes,
            ),
            total=len(texts),
            desc="Processing",
        )
    )

    return processed_texts



from .hf_local import HuggingFaceLocalTokenizer


class HuggingFaceRemoteTokenizer(HuggingFaceLocalTokenizer):
    """
    Tokenizer for HuggingFace models served in Inference API or Endpoints.
    """

    def __init__(self, model_name: str):
        super().__init__(model_name)



from abc import ABC, abstractmethod
from typing import Dict, List


class Tokenizer(ABC):
    """
    Base class for tokenizers.
    """

    def __init__(self, model_name: str):
        self.model_name = model_name

    @abstractmethod
    def encode(self, text: str) -> List[int]:
        """
        Encodes a text in token ids.

        Args:
            text (str): a text.

        Returns:
            List[int]: list of token ids.
        """
        ...

    @abstractmethod
    def decode(self, tokens: List[int]) -> str:
        """
        Decodes a list of token ids.

        Args:
            tokens (List[int]): list of token ids.

        Returns:
            text (str): decoded text.
        """
        ...

    def get_token_length(self, text: str) -> int:
        """
        Get the token length of a text.

        Args:
            text (str): a text.

        Returns:
            int: token length of the text.
        """
        return len(self.encode(text))

    def truncate_text(self, text: str, max_tokens: int) -> str:
        """
        Truncates a text to a maximum token length.

        Args:
            text (str): a text.
            max_tokens (int): max token length of the text after truncating.

        Returns:
            str: truncated text.
        """
        return self.decode(self.encode(text)[:max_tokens])

    def truncate_texts(self, texts: List[str], max_tokens: int) -> List[str]:
        """
        Truncates a list of texts to a maximum token length.

        Args:
            texts (List[str]): a list of texts.
            max_tokens (int): max token length of each text after truncating.

        Returns:
            List[str]: list of truncated text.
        """
        return [self.truncate_text(text, max_tokens) for text in texts]

    def distributed_truncate(
        self, texts: Dict[str, List[str]], max_tokens: int
    ) -> Dict[str, List[str]]:
        """
        Truncates texts from different extractors to a maximum token length.
        It distributes the `max_tokens` across all the extractor keys,
        so, when all of them are included in the prompt, they sum at most
        `max_tokens`.

        Example:
            texts = {"summary": ["A", "B"], "headline": ["C", "D"]}
            max_tokens = 256
            output = {
                "summary": [truncated("A", 128), truncated("B", 128)],
                "headline": [truncated("C", 128), truncated("D", 128)],
            }

        Args:
            texts (Dict[str, List[str]]): texts of each extractor.
            max_tokens (int): max length to be distributed across extractors.

        Returns:
            Dict[str, List[str]]: truncated texts of each extractor.
        """
        max_tokens = max_tokens // len(texts)
        for key in texts:
            texts[key] = self.truncate_texts(texts[key], max_tokens)

        return texts



import os
from typing import Dict

import requests
from requests.adapters import HTTPAdapter, Retry
from transformers import AutoTokenizer

from ..common.logging import get_logger
from ..config import ModelConfig
from .base import TextGenerationModel
from .types import GENERATION_ERROR, CompletionType

_logger = get_logger(__name__)


class HuggingFaceRemoteModel(TextGenerationModel):
    """
    Generates completions using HuggingFace's models remotely deployed
    (HuggingFace's Inference API or Inference Endpoints).

    Requires the definition of the `HF_TOKEN=<token>` environment variable.
    """

    def __init__(self, model_config: ModelConfig):
        super().__init__(model_config)
        self.client = requests.Session()
        retry_adapter = HTTPAdapter(
            max_retries=Retry(
                total=getattr(self.model_config, "max_retries", 5),
                backoff_factor=getattr(
                    self.model_config, "backoff_factor", 0.5
                ),
                status_forcelist=[
                    code for code in requests.status_codes._codes if code != 200
                ],
            )
        )
        self.client.mount("http://", retry_adapter)
        self.client.mount("https://", retry_adapter)

        if self.model_config.api_type == CompletionType.CHAT:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.model_config.model_name
            )

    def generate_completion(self, prompt: str, generation_config: Dict) -> str:
        headers = {"Authorization": f'Bearer {os.environ["HF_TOKEN"]}'}

        inputs = prompt
        if self.model_config.api_type == CompletionType.CHAT:
            inputs = self.tokenizer.apply_chat_template(
                [
                    {"role": "user", "content": prompt},
                ],
                add_generation_prompt=True,
                tokenize=False,
            )

        payload = {
            "inputs": inputs,
            "parameters": generation_config,
        }
        try:
            response = self.client.post(
                self.model_config.url, headers=headers, json=payload
            )
            return response.json()[0]["generated_text"]
        except Exception as e:
            _logger.info(f"Unrecoverable exception during the request: {e}")
            return GENERATION_ERROR



from typing import Dict, Final

EXTRACTOR_OMITTED: Final = "<omitted>"

SPACY_MODEL_MAPPING: Final[Dict[str, str]] = {
    "ca": "ca_core_news_sm",
    "zh": "zh_core_web_sm",
    "hr": "hr_core_news_sm",
    "da": "da_core_news_sm",
    "nl": "nl_core_news_sm",
    "en": "en_core_web_sm",
    "fi": "fi_core_news_sm",
    "fr": "fr_core_news_sm",
    "de": "de_core_news_sm",
    "el": "el_core_news_sm",
    "it": "it_core_news_sm",
    "ja": "ja_core_news_sm",
    "ko": "ko_core_news_sm",
    "lt": "lt_core_news_sm",
    "mk": "mk_core_news_sm",
    "nb": "nb_core_news_sm",
    "pl": "pl_core_news_sm",
    "pt": "pt_core_news_sm",
    "ro": "ro_core_news_sm",
    "ru": "ru_core_news_sm",
    "sl": "sl_core_news_sm",
    "es": "es_core_news_sm",
    "sv": "sv_core_news_sm",
    "uk": "uk_core_news_sm",
    "multilingual": "xx_sent_ud_sm",
}



# flake8: noqa
from importlib import import_module
from typing import Mapping

from ..common.exceptions import MissingIntegrationError
from ..common.logging import get_logger
from .base import Tokenizer

_logger = get_logger(__name__)


TOKENIZERS: Mapping[str, str] = {
    "openai": "OpenAITokenizer",
    "anthropic": "AnthropicTokenizer",
    "cohere": "CohereTokenizer",
    "hf_local": "HuggingFaceLocalTokenizer",
    "hf_remote": "HuggingFaceRemoteTokenizer",
    "vertex": "VertexTokenizer",
    "bedrock": "BedrockTokenizer",
    "ai21": "AI21Tokenizer",
    "azure_openai": "AzureOpenAITokenizer",
    "inference_server": "InferenceServerTokenizer",
}


def get_tokenizer(provider: str, model_name: str) -> Tokenizer:
    """
    Gets a tokenizer from the pool.

    Args:
        provider (str): a model provider.
        model_name (str): name of a model served by the provider.

    Returns:
        Tokenizer: a tokenizer from the pool.
    """
    tokenizer_cls_name = TOKENIZERS[provider]

    if tokenizer_cls_name == "vertex":
        _logger.warn(
            "Vertex does not provide a tokenizer,"
            "OpenAI tokenizer for GPT-4 will be used."
        )

    try:
        tokenizer_class = getattr(
            import_module(f".{provider}", __name__),
            tokenizer_cls_name,
        )
    except (ModuleNotFoundError, ImportError):
        raise MissingIntegrationError(integration=provider)

    return tokenizer_class(model_name)


__all__ = list(TOKENIZERS.values())



import importlib
from functools import reduce
from typing import Dict, List

from datasets import Dataset

from ..config import InputConfig
from ..types import TaskType
from .base import Extractor


class Combined(Extractor):
    """
    Extractor that combines multiple extractors.

    This extractor does not need specific template placeholders,
    just the placehoders of the extractors being combined.

    This extractor does not need specific arguments,
    just the arguments for the extractors being combined.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        super().__init__(input_config, task_type)
        self.extractors = []
        for extractor_name in self.input_config.extractors_list:
            module, class_ = extractor_name.rsplit(".", 1)
            extractor = getattr(
                importlib.import_module(
                    "." + module, package=__name__.rsplit(".", 1)[0]
                ),
                class_,
            )(self.input_config, task_type)
            self.extractors.append(extractor)

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        extractor_outputs = [
            extractor._extract(dataset) for extractor in self.extractors
        ]
        return reduce(lambda x, y: dict(x, **y), extractor_outputs)



from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from typing import Dict, List

from tqdm import tqdm

from ..config import ModelConfig


class TextGenerationModel(ABC):
    """
    Base class for LLMs.
    """

    def __init__(self, model_config: ModelConfig):
        self.model_config = deepcopy(model_config)

    @abstractmethod
    def generate_completion(
        self,
        prompt: str,
        generation_config: Dict,
    ) -> str:
        """
        Generates a completion for a `prompt` by decoding a model
        parameterized by `generation_config`. This method has to be
        overwritten to implement the completion code.

        Args:
            prompts (str): prompt to generate completions for.
            generation_config (Dict): Dictionary containing the generation parameters.

        Returns:
            str: Generated completion or `.types.GENERATION_ERROR` if there was some error.
        """
        ...

    def generate_completions(
        self,
        prompts: List[str],
        generation_config: Dict,
    ) -> List[str]:
        """Generates a completion for each prompt in a list of `prompts`.

        Args:
            prompts (List[str]): List of prompts to generate completions for.
            generation_config (Dict): Dictionary containing the generation parameters.

        Returns:
            List[str]: List of generated completions.
        """
        completions, responses = [], []
        with ThreadPoolExecutor(
            max_workers=min(self.model_config.threads, len(prompts))
        ) as thread_pool:
            for prompt in prompts:
                responses.append(
                    thread_pool.submit(
                        self.generate_completion, prompt, generation_config
                    )
                )
            # Wait completions
            completions = [response.result() for response in tqdm(responses)]

        return completions



import os
from typing import Dict

from openai import OpenAI

from ..common.logging import get_logger
from ..common.utils import get_instantiation_args
from ..config import ModelConfig
from .base import TextGenerationModel
from .types import GENERATION_ERROR, CompletionType

_logger = get_logger(__name__)


class OpenAIModel(TextGenerationModel):
    """
    Generates completions using OpenAI models.

    Requires the definition of the `OPENAI_API_KEY=<key>` environment variable.
    """

    def __init__(self, model_config: ModelConfig):
        super().__init__(model_config)
        self.client = OpenAI(
            api_key=os.environ["OPENAI_API_KEY"],
            **get_instantiation_args(
                OpenAI.__init__, self.model_config.model_dump()
            ),
        )

    def generate_completion(
        self,
        prompt: str,
        generation_config: Dict,
    ) -> str:
        completion_fn = (
            self._chat_request
            if self.model_config.api_type == CompletionType.CHAT
            else self._completion_request
        )
        try:
            completion = completion_fn(prompt, generation_config)
        except Exception as e:
            _logger.info(f"Unrecoverable exception during the request: {e}")
            return GENERATION_ERROR
        return completion

    def _chat_request(self, prompt: str, generation_config: Dict) -> str:
        return (
            self.client.chat.completions.create(
                messages=[
                    {
                        "role": "user",
                        "content": prompt,
                    }
                ],
                model=self.model_config.model_name,
                **generation_config,
            )
            .choices[0]
            .message.content
        )

    def _completion_request(self, prompt: str, generation_config: Dict) -> str:
        return (
            self.client.completions.create(
                prompt=prompt,
                model=self.model_config.model_name,
                **generation_config,
            )
            .choices[0]
            .text
        )



from abc import ABC, abstractmethod
from typing import Any, Dict, List

from datasets import Dataset

from ..common.exceptions import ExtractorEmptyColumns
from ..config import InputConfig
from ..types import TaskType
from .utils import clean_inputs


class Extractor(ABC):
    """
    Base class for an extractor.
    """

    def __init__(
        self,
        input_config: InputConfig,
        task_type: TaskType,
        workspace: Dict[str, Any] = {},
        args: Dict[str, Any] = {},
    ):
        self.input_config = input_config
        self.task_type = task_type
        self.workspace: Dict[str, Any] = workspace
        self.args: Dict[str, Any] = args
        self.check_valid_args()

    def check_valid_args(self) -> None:
        """
        Checks if the arguments passed to the extractor are valid.

        Raises:
            ExtractorInvalidArgs: if the arguments are invalid.

        """
        ...

    @abstractmethod
    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        """
        Returns the prompt inputs for each sample in a dataset.
        This method must be overridden in each new extractor.

        Example:
            input = Dataset({"text": ["hi Jose", "hi Areg"], "label": [0, 1]})
            output = {"entities": ["Jose", "Areg"], "interject": ["hi", "hi"]}

        Args:
            dataset (Dataset): A dataset to extract inputs from.

        Returns:
            Dict[str, List[str]]: A dictionary mapping each template
                                  key to a list of prompt inputs
                                  (one input per template key and example).
        """
        ...

    def prepare_human(self, human_texts: List[str]) -> List[str]:
        """
        Prepares the human texts. Some extractors could need to modify
        human texts according to the extractions, e.g., remove prefixes
        from texts to ensure that generations and human texts are
        continuations of the same prefix.

        Args:
            human_texts (List[str]): list of human texts.

        Returns:
            List[str]: prepared human texts.
        """
        return human_texts

    def extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        """
        Calls _extract and cleans the extracted inputs.

        Args:
            dataset (Dataset): A dataset to extract inputs from.

        Returns:
            Dict[str, List[str]]: A dictionary mapping each template
                                  key to a list of prompt inputs
                                  (one input per template key and example).

        Raises:
            ExtractorEmptyColumns: if any field of the prompt_inputs is empty.
        """
        prompt_inputs = self._extract(dataset)
        prompt_inputs = {
            column: clean_inputs(prompt_inputs[column])
            for column in prompt_inputs
        }

        for field in prompt_inputs:
            if len(prompt_inputs[field]) == 0:
                raise ExtractorEmptyColumns(self.__class__.__name__, field)

        return prompt_inputs
